@@ -1,4 +1,4 @@
-from typing import Union, List, Optional
+from typing import List, Optional
 import time
 import gymnasium as gym
 import os
@@ -24,6 +24,7 @@ from csle_common.dao.simulation_config.state import State
 from csle_common.dao.simulation_config.action import Action
 from csle_common.dao.training.player_type import PlayerType
 from csle_common.util.general_util import GeneralUtil
+from csle_common.dao.simulation_config.base_env import BaseEnv
 from csle_agents.agents.base.base_agent import BaseAgent
 import csle_agents.constants.constants as agents_constants
 
@@ -33,15 +34,30 @@ class DQNAgent(BaseAgent):
     A DQN agent using the implementation from OpenAI baselines
     """
     def __init__(self, simulation_env_config: SimulationEnvConfig,
-                 emulation_env_config: Union[None, EmulationEnvConfig], experiment_config: ExperimentConfig,
-                 training_job: Optional[TrainingJobConfig] = None):
+                 emulation_env_config: Optional[EmulationEnvConfig], experiment_config: ExperimentConfig,
+                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True) -> None:
+        """
+        Initializes the agent
+
+        :param simulation_env_config: the simulation environment configuration
+        :param emulation_env_config: the emulation environment configuration
+        :param experiment_config: the experiment configuration
+        :param training_job: the training job configuration
+        :param save_to_metastore: boolean flag indicating whether job information should be saved to the metastore
+        """
         super(DQNAgent, self).__init__(simulation_env_config=simulation_env_config,
                                        emulation_env_config=emulation_env_config,
                                        experiment_config=experiment_config)
         assert experiment_config.agent_type == AgentType.DQN
         self.training_job = training_job
+        self.save_to_metastore = save_to_metastore
 
     def train(self) -> ExperimentExecution:
+        """
+        Implements the training logic of the DQN algorithm
+
+        :return: the experiment result
+        """
         pid = os.getpid()
 
         # Setup experiment metrics
@@ -52,10 +68,13 @@ class DQNAgent(BaseAgent):
 
         # Setup training job
         if self.training_job is None:
+            emulation_name = ""
+            if self.emulation_env_config is not None:
+                emulation_name = self.emulation_env_config.name
             self.training_job = TrainingJobConfig(
                 simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                 progress_percentage=0, pid=pid, experiment_result=exp_result,
-                emulation_env_name=self.emulation_env_config.name, simulation_traces=[],
+                emulation_env_name=emulation_name, simulation_traces=[],
                 num_cached_traces=agents_constants.COMMON.NUM_CACHED_SIMULATION_TRACES,
                 log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                 physical_host_ip=GeneralUtil.get_host_ip())
@@ -69,10 +88,10 @@ class DQNAgent(BaseAgent):
 
         # Setup experiment execution
         ts = time.time()
-        emulation_name = None
+        simulation_name = self.simulation_env_config.name
+        emulation_name = ""
         if self.emulation_env_config is not None:
             emulation_name = self.emulation_env_config.name
-        simulation_name = self.simulation_env_config.name
         self.exp_execution = ExperimentExecution(
             result=exp_result, config=self.experiment_config, timestamp=ts,
             emulation_name=emulation_name, simulation_name=simulation_name,
@@ -82,7 +101,7 @@ class DQNAgent(BaseAgent):
 
         # Setup gym environment
         config = self.simulation_env_config.simulation_env_input_config
-        orig_env = gym.make(self.simulation_env_config.gym_env_name, config=config)
+        orig_env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=config)
         env = make_vec_env(env_id=self.simulation_env_config.gym_env_name,
                            n_envs=self.experiment_config.hparams[agents_constants.COMMON.NUM_PARALLEL_ENVS].value,
                            env_kwargs={"config": config}, vec_env_cls=DummyVecEnv)
@@ -136,7 +155,7 @@ class DQNAgent(BaseAgent):
                 buffer_size=self.experiment_config.hparams[agents_constants.DQN.BUFFER_SIZE].value
             )
             if self.experiment_config.player_type == PlayerType.ATTACKER:
-                orig_env.set_model(model)
+                orig_env.unwrapped.set_model(model)
 
             # Train
             model.learn(total_timesteps=self.experiment_config.hparams[
@@ -161,8 +180,8 @@ class DQNAgent(BaseAgent):
             os.chmod(save_path, 0o777)
 
             # Save latest trace
-            MetastoreFacade.save_simulation_trace(orig_env.get_traces()[-1])
-            orig_env.reset_traces()
+            MetastoreFacade.save_simulation_trace(orig_env.unwrapped.get_traces()[-1])
+            orig_env.unwrapped.reset_traces()
 
         # Calculate average and std metrics
         exp_result.avg_metrics = {}
@@ -187,13 +206,15 @@ class DQNAgent(BaseAgent):
             exp_result.avg_metrics[metric] = avg_metrics
             exp_result.std_metrics[metric] = std_metrics
 
-        traces = orig_env.get_traces()
+        traces = orig_env.unwrapped.get_traces()
         if len(traces) > 0:
             MetastoreFacade.save_simulation_trace(traces[-1])
         return self.exp_execution
 
     def hparam_names(self) -> List[str]:
         """
+        Gets the hyperparameters
+
         :return: a list with the hyperparameter names
         """
         return [constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER,
@@ -219,7 +240,7 @@ class DQNTrainingCallback(BaseCallback):
                  training_job: TrainingJobConfig, exp_execution: ExperimentExecution,
                  max_steps: int, simulation_name: str,
                  states: List[State], actions: List[Action], player_type: PlayerType,
-                 env: gym.Env, experiment_config: ExperimentConfig, verbose=0,
+                 env: BaseEnv, experiment_config: ExperimentConfig, verbose=0,
                  eval_every: int = 100, eval_batch_size: int = 10, save_every: int = 10, save_dir: str = "",
                  L: int = 3, gym_env_name: str = ""):
         """
@@ -261,7 +282,7 @@ class DQNTrainingCallback(BaseCallback):
         self.actions = actions
         self.save_every = save_every
         self.save_dir = save_dir
-        self.env = env
+        self.env: BaseEnv = env
         self.experiment_config = experiment_config
         self.L = L
         self.gym_env_name = gym_env_name
@@ -290,7 +311,7 @@ class DQNTrainingCallback(BaseCallback):
         :return: (bool) If the callback returns False, training is aborted early.
         """
         if self.player_type == PlayerType.ATTACKER:
-            self.env.set_model(self.model)
+            self.env.unwrapped.set_model(self.model)
         return True
 
     def _on_training_end(self) -> None:
@@ -318,22 +339,22 @@ class DQNTrainingCallback(BaseCallback):
         # Eval model
         if self.iter % self.eval_every == 0:
             if self.player_type == PlayerType.ATTACKER:
-                self.env.set_model(self.model)
+                self.env.unwrapped.set_model(self.model)
             policy = DQNPolicy(
                 model=self.model, simulation_name=self.simulation_name, save_path=save_path,
                 states=self.states, player_type=self.player_type, actions=self.actions,
                 experiment_config=self.experiment_config, avg_R=-1)
-            o = self.env.reset()
+            o, _ = self.env.reset()
             max_horizon = self.experiment_config.hparams[agents_constants.COMMON.MAX_ENV_STEPS].value
             avg_rewards = []
             for i in range(self.eval_batch_size):
-                o = self.env.reset()
+                o, _ = self.env.reset()
                 done = False
                 t = 0
                 cumulative_reward = 0
                 while not done and t <= max_horizon:
                     a = policy.action(o=o)
-                    o, r, done, info = self.env.step(a)
+                    o, r, done, _, info = self.env.step(a)
                     cumulative_reward += r
                     t += 1
                     Logger.__call__().get_logger().debug(f"t:{t}, a1:{a}, r:{r}, info:{info}, done:{done}")
@@ -352,7 +373,8 @@ class DQNTrainingCallback(BaseCallback):
             progress = round(steps_done / total_steps_done, 2)
             self.training_job.progress_percentage = progress
             self.training_job.experiment_result = self.exp_result
-            self.training_job.simulation_traces.append(self.env.get_traces()[-1])
+            if len(self.env.unwrapped.get_traces()) > 0:
+                self.training_job.simulation_traces.append(self.env.unwrapped.get_traces()[-1])
             if len(self.training_job.simulation_traces) > self.training_job.num_cached_traces:
                 self.training_job.simulation_traces = self.training_job.simulation_traces[1:]
             MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)
