@@ -1,6 +1,6 @@
 from typing import Union, List, Optional
 import time
-import gymnasium as gym
+import gymnasium
 import os
 import numpy as np
 import math
@@ -25,6 +25,7 @@ from csle_common.dao.simulation_config.state import State
 from csle_common.dao.simulation_config.action import Action
 from csle_common.dao.training.player_type import PlayerType
 from csle_common.util.general_util import GeneralUtil
+from csle_common.dao.simulation_config.base_env import BaseEnv
 from csle_agents.agents.base.base_agent import BaseAgent
 import csle_agents.constants.constants as agents_constants
 
@@ -36,14 +37,29 @@ class PPOAgent(BaseAgent):
 
     def __init__(self, simulation_env_config: SimulationEnvConfig,
                  emulation_env_config: Union[None, EmulationEnvConfig], experiment_config: ExperimentConfig,
-                 training_job: Optional[TrainingJobConfig] = None):
+                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True):
+        """
+        Intializes the agent
+
+        :param simulation_env_config: the simulation environment configuration
+        :param emulation_env_config: the emulation environment configuration
+        :param experiment_config: the experiment configuration
+        :param training_job: the training job
+        :param save_to_metastore: boolean flag indicating whether the results should be saved to the metastore or not
+        """
         super(PPOAgent, self).__init__(simulation_env_config=simulation_env_config,
                                        emulation_env_config=emulation_env_config,
                                        experiment_config=experiment_config)
         assert experiment_config.agent_type == AgentType.PPO
         self.training_job = training_job
+        self.save_to_metastore = save_to_metastore
 
     def train(self) -> ExperimentExecution:
+        """
+        Runs the training process
+
+        :return: the results
+        """
         pid = os.getpid()
 
         # Setup experiment metrics
@@ -54,30 +70,37 @@ class PPOAgent(BaseAgent):
         exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_TIME_HORIZON)
         exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN)
         exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_RANDOM_RETURN)
+        exp_result.plot_metrics.append(agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN)
         exp_result.plot_metrics.append(agents_constants.COMMON.RUNTIME)
         descr = f"Training of policies with PPO using " \
                 f"simulation:{self.simulation_env_config.name}"
 
         # Setup training job
         if self.training_job is None:
+            emulation_name = ""
+            if self.emulation_env_config is not None:
+                emulation_name = self.emulation_env_config.name
             self.training_job = TrainingJobConfig(
                 simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                 progress_percentage=0, pid=pid, experiment_result=exp_result,
-                emulation_env_name=self.emulation_env_config.name, simulation_traces=[],
+                emulation_env_name=emulation_name, simulation_traces=[],
                 num_cached_traces=agents_constants.COMMON.NUM_CACHED_SIMULATION_TRACES,
                 log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                 physical_host_ip=GeneralUtil.get_host_ip())
-            training_job_id = MetastoreFacade.save_training_job(training_job=self.training_job)
+            training_job_id = -1
+            if self.save_to_metastore:
+                training_job_id = MetastoreFacade.save_training_job(training_job=self.training_job)
             self.training_job.id = training_job_id
         else:
             self.training_job.pid = pid
             self.training_job.progress_percentage = 0
             self.training_job.experiment_result = exp_result
-            MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)
+            if self.save_to_metastore:
+                MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)
 
         # Setup experiment execution
         ts = time.time()
-        emulation_name = None
+        emulation_name = ""
         if self.emulation_env_config is not None:
             emulation_name = self.emulation_env_config.name
         simulation_name = self.simulation_env_config.name
@@ -85,12 +108,14 @@ class PPOAgent(BaseAgent):
             result=exp_result, config=self.experiment_config, timestamp=ts,
             emulation_name=emulation_name, simulation_name=simulation_name, descr=descr,
             log_file_path=self.training_job.log_file_path)
-        exp_execution_id = MetastoreFacade.save_experiment_execution(self.exp_execution)
+        exp_execution_id = -1
+        if self.save_to_metastore:
+            exp_execution_id = MetastoreFacade.save_experiment_execution(self.exp_execution)
         self.exp_execution.id = exp_execution_id
 
         # Setup gym environment
         config = self.simulation_env_config.simulation_env_input_config
-        orig_env = gym.make(self.simulation_env_config.gym_env_name, config=config)
+        orig_env: BaseEnv = gymnasium.make(self.simulation_env_config.gym_env_name, config=config)
         env = make_vec_env(env_id=self.simulation_env_config.gym_env_name,
                            n_envs=self.experiment_config.hparams[agents_constants.COMMON.NUM_PARALLEL_ENVS].value,
                            env_kwargs={"config": config}, vec_env_cls=DummyVecEnv)
@@ -98,7 +123,7 @@ class PPOAgent(BaseAgent):
 
         # Training runs, one per seed
         for seed in self.experiment_config.random_seeds:
-            self.start = time.time()
+            self.start: float = time.time()
             exp_result.all_metrics[seed] = {}
             exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = []
             exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = []
@@ -106,10 +131,14 @@ class PPOAgent(BaseAgent):
             exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON] = []
             exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN] = []
             exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RANDOM_RETURN] = []
+            exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN] = []
             exp_result.all_metrics[seed][agents_constants.COMMON.RUNTIME] = []
             ExperimentUtil.set_seed(seed)
 
             # Callback for logging training metrics
+            L = 1
+            if agents_constants.COMMON.L in self.experiment_config.hparams:
+                L = self.experiment_config.hparams[agents_constants.COMMON.L].value
             cb = PPOTrainingCallback(
                 eval_every=self.experiment_config.hparams[agents_constants.COMMON.EVAL_EVERY].value,
                 eval_batch_size=self.experiment_config.hparams[agents_constants.COMMON.EVAL_BATCH_SIZE].value,
@@ -124,9 +153,9 @@ class PPOAgent(BaseAgent):
                 save_every=self.experiment_config.hparams[agents_constants.COMMON.SAVE_EVERY].value,
                 save_dir=self.experiment_config.output_dir, exp_execution=self.exp_execution,
                 env=orig_env, experiment_config=self.experiment_config,
-                L=self.experiment_config.hparams[agents_constants.COMMON.L].value,
+                L=L,
                 gym_env_name=self.simulation_env_config.gym_env_name,
-                start=self.start
+                start=self.start, save_to_metastore=self.save_to_metastore
             )
 
             # Create PPO Agent
@@ -147,10 +176,11 @@ class PPOAgent(BaseAgent):
                 vf_coef=self.experiment_config.hparams[agents_constants.PPO.VF_COEF].value,
                 max_grad_norm=self.experiment_config.hparams[agents_constants.PPO.MAX_GRAD_NORM].value,
                 target_kl=self.experiment_config.hparams[agents_constants.PPO.TARGET_KL].value,
+                n_epochs=self.experiment_config.hparams[agents_constants.PPO.NUM_GRADIENT_STEPS].value
             )
             if self.experiment_config.player_type == PlayerType.ATTACKER \
                     and "stopping" in self.simulation_env_config.gym_env_name:
-                orig_env.set_model(model)
+                orig_env.unwrapped.set_model(model)
 
             # Train
             model.learn(total_timesteps=self.experiment_config.hparams[
@@ -171,12 +201,14 @@ class PPOAgent(BaseAgent):
             exp_result.policies[seed] = policy
 
             # Save policy metadata
-            MetastoreFacade.save_ppo_policy(ppo_policy=policy)
-            os.chmod(save_path, 0o777)
+            if self.save_to_metastore:
+                MetastoreFacade.save_ppo_policy(ppo_policy=policy)
+                os.chmod(save_path, 0o777)
 
             # Save latest trace
-            MetastoreFacade.save_simulation_trace(orig_env.get_traces()[-1])
-            orig_env.reset_traces()
+            if self.save_to_metastore and len(orig_env.unwrapped.get_traces()) > 0:
+                MetastoreFacade.save_simulation_trace(orig_env.unwrapped.get_traces()[-1])
+            orig_env.unwrapped.reset_traces()
 
         # Calculate average and std metrics
         exp_result.avg_metrics = {}
@@ -201,8 +233,8 @@ class PPOAgent(BaseAgent):
             exp_result.avg_metrics[metric] = avg_metrics
             exp_result.std_metrics[metric] = std_metrics
 
-        traces = orig_env.get_traces()
-        if len(traces) > 0:
+        traces = orig_env.unwrapped.get_traces()
+        if len(traces) > 0 and self.save_to_metastore:
             MetastoreFacade.save_simulation_trace(traces[-1])
         return self.exp_execution
 
@@ -217,9 +249,10 @@ class PPOAgent(BaseAgent):
                 agents_constants.COMMON.GAMMA, agents_constants.PPO.GAE_LAMBDA, agents_constants.PPO.CLIP_RANGE,
                 agents_constants.PPO.CLIP_RANGE_VF, agents_constants.PPO.ENT_COEF,
                 agents_constants.PPO.VF_COEF, agents_constants.PPO.MAX_GRAD_NORM, agents_constants.PPO.TARGET_KL,
+                agents_constants.PPO.NUM_GRADIENT_STEPS,
                 agents_constants.COMMON.NUM_TRAINING_TIMESTEPS, agents_constants.COMMON.EVAL_EVERY,
                 agents_constants.COMMON.EVAL_BATCH_SIZE, constants.NEURAL_NETWORKS.DEVICE,
-                agents_constants.COMMON.SAVE_EVERY]
+                agents_constants.COMMON.SAVE_EVERY, agents_constants.COMMON.EVALUATE_WITH_DISCOUNT]
 
 
 class PPOTrainingCallback(BaseCallback):
@@ -231,9 +264,9 @@ class PPOTrainingCallback(BaseCallback):
                  training_job: TrainingJobConfig, exp_execution: ExperimentExecution,
                  max_steps: int, simulation_name: str, start: float,
                  states: List[State], actions: List[Action], player_type: PlayerType,
-                 env: gym.Env, experiment_config: ExperimentConfig, verbose=0,
+                 env: BaseEnv, experiment_config: ExperimentConfig, verbose=0,
                  eval_every: int = 100, eval_batch_size: int = 10, save_every: int = 10, save_dir: str = "",
-                 L: int = 3, gym_env_name: str = ""):
+                 L: int = 3, gym_env_name: str = "", save_to_metastore: bool = False):
         """
         Initializes the callback
 
@@ -257,6 +290,7 @@ class PPOTrainingCallback(BaseCallback):
         :param L: num stops if a stopping environment
         :param gym_env_name: name of gym env
         :param start_time: the start time-stamp
+        :param save_to_metastore: boolean flag indicating whether the results should be saved to the metastore
         """
         super(PPOTrainingCallback, self).__init__(verbose)
         self.states = states
@@ -274,11 +308,12 @@ class PPOTrainingCallback(BaseCallback):
         self.actions = actions
         self.save_every = save_every
         self.save_dir = save_dir
-        self.env = env
+        self.env: BaseEnv = env
         self.experiment_config = experiment_config
         self.L = L
         self.gym_env_name = gym_env_name
         self.start = start
+        self.save_to_metastore = save_to_metastore
 
     def _on_training_start(self) -> None:
         """
@@ -305,7 +340,7 @@ class PPOTrainingCallback(BaseCallback):
         """
         if self.experiment_config.player_type == PlayerType.ATTACKER \
                 and "stopping" in self.simulation_name:
-            self.env.set_model(self.model)
+            self.env.unwrapped.set_model(self.model)
         return True
 
     def _on_training_end(self) -> None:
@@ -329,32 +364,44 @@ class PPOTrainingCallback(BaseCallback):
             Logger.__call__().get_logger().info(f"Saving model to path: {save_path}")
             self.model.save(save_path)
             os.chmod(save_path, 0o777)
+            policy = PPOPolicy(
+                model=self.model, simulation_name=self.simulation_name, save_path=save_path,
+                states=[],
+                actions=[], player_type=self.experiment_config.player_type,
+                experiment_config=self.experiment_config,
+                avg_R=self.exp_result.all_metrics[self.seed][agents_constants.COMMON.AVERAGE_RETURN][-1])
+            self.exp_result.policies[self.seed] = policy
+            MetastoreFacade.save_ppo_policy(ppo_policy=policy)
 
         # Eval model
         if self.iter % self.eval_every == 0:
             if self.player_type == PlayerType.ATTACKER and "stopping" in self.simulation_name:
-                self.env.set_model(self.model)
+                self.env.unwrapped.set_model(self.model)
             policy = PPOPolicy(
                 model=self.model, simulation_name=self.simulation_name, save_path=save_path,
                 states=self.states, player_type=self.player_type, actions=self.actions,
                 experiment_config=self.experiment_config, avg_R=-1)
-            o = self.env.reset()
+            o, _ = self.env.reset()
             max_horizon = self.experiment_config.hparams[agents_constants.COMMON.MAX_ENV_STEPS].value
             avg_rewards = []
             avg_horizons = []
             avg_upper_bounds = []
             avg_random_returns = []
+            avg_heuristic_returns = []
             info = {}
             for i in range(self.eval_batch_size):
-                o = self.env.reset()
+                o, _ = self.env.reset()
                 done = False
                 t = 0
                 cumulative_reward = 0
                 while not done and t <= max_horizon:
                     a = policy.action(o=o)
-                    o, r, done, info = self.env.step(a)
-                    cumulative_reward += r * math.pow(
-                        self.experiment_config.hparams[agents_constants.COMMON.GAMMA].value, t)
+                    o, r, done, _, info = self.env.step(a)
+                    if self.experiment_config.hparams[agents_constants.COMMON.EVALUATE_WITH_DISCOUNT].value:
+                        cumulative_reward += r * math.pow(
+                            self.experiment_config.hparams[agents_constants.COMMON.GAMMA].value, t)
+                    else:
+                        cumulative_reward += r
                     t += 1
                     Logger.__call__().get_logger().debug(f"t:{t}, a1:{a}, r:{r}, info:{info}, done:{done}")
                 avg_rewards.append(cumulative_reward)
@@ -370,14 +417,21 @@ class PPOTrainingCallback(BaseCallback):
                     avg_random_returns.append(info[agents_constants.ENV_METRICS.AVERAGE_RANDOM_RETURN])
                 else:
                     avg_random_returns.append(-1)
-
-                avg_upper_bounds.append(info[agents_constants.ENV_METRICS.AVERAGE_UPPER_BOUND_RETURN])
+                if agents_constants.ENV_METRICS.AVERAGE_HEURISTIC_RETURN in info:
+                    avg_heuristic_returns.append(info[agents_constants.ENV_METRICS.AVERAGE_HEURISTIC_RETURN])
+                else:
+                    avg_heuristic_returns.append(-1)
+                if agents_constants.ENV_METRICS.AVERAGE_UPPER_BOUND_RETURN in info:
+                    avg_upper_bounds.append(info[agents_constants.ENV_METRICS.AVERAGE_UPPER_BOUND_RETURN])
+                else:
+                    avg_upper_bounds.append(-1)
             avg_R = np.mean(avg_rewards)
             avg_T = np.mean(avg_horizons)
             avg_random_return = np.mean(avg_random_returns)
+            avg_heuristic_return = np.mean(avg_heuristic_returns)
             avg_upper_bound = np.mean(avg_upper_bounds)
             policy.avg_R = avg_R
-            time_elapsed_minutes = (time.time() - self.start) // 60
+            time_elapsed_minutes = round((time.time() - self.start) / 60, 3)
             self.exp_result.all_metrics[self.seed][agents_constants.COMMON.AVERAGE_RETURN].append(round(avg_R, 3))
             self.exp_result.all_metrics[self.seed][agents_constants.COMMON.AVERAGE_TIME_HORIZON].append(round(avg_T, 3))
             self.exp_result.all_metrics[self.seed][agents_constants.COMMON.AVERAGE_UPPER_BOUND_RETURN].append(
@@ -385,6 +439,8 @@ class PPOTrainingCallback(BaseCallback):
             self.exp_result.all_metrics[self.seed][agents_constants.COMMON.RUNTIME].append(time_elapsed_minutes)
             self.exp_result.all_metrics[self.seed][agents_constants.COMMON.AVERAGE_RANDOM_RETURN].append(
                 round(avg_random_return, 3))
+            self.exp_result.all_metrics[self.seed][agents_constants.COMMON.AVERAGE_HEURISTIC_RETURN].append(
+                round(avg_heuristic_return, 3))
             running_avg_J = ExperimentUtil.running_average(
                 self.exp_result.all_metrics[self.seed][agents_constants.COMMON.AVERAGE_RETURN],
                 self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
@@ -401,7 +457,8 @@ class PPOTrainingCallback(BaseCallback):
                 f"{round(running_avg_J, 3)}, Avg T:{round(avg_T, 3)}, "
                 f"Running_avg_{self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value}_T: "
                 f"{round(running_avg_T, 3)}, Avg pi*: {round(avg_upper_bound, 3)}, "
-                f"Avg random R:{round(avg_random_return, 3)}, time elapsed (min): {time_elapsed_minutes}")
+                f"Avg random R:{round(avg_random_return, 3)}, Avg heuristic R:{round(avg_heuristic_return, 3)}, "
+                f"time elapsed (min): {time_elapsed_minutes}")
 
             self.env.reset()
 
@@ -411,17 +468,19 @@ class PPOTrainingCallback(BaseCallback):
             progress = round(steps_done / total_steps_done, 2)
             self.training_job.progress_percentage = progress
             self.training_job.experiment_result = self.exp_result
-            if len(self.env.get_traces()) > 0:
-                self.training_job.simulation_traces.append(self.env.get_traces()[-1])
+            if len(self.env.unwrapped.get_traces()) > 0:
+                self.training_job.simulation_traces.append(self.env.unwrapped.get_traces()[-1])
             if len(self.training_job.simulation_traces) > self.training_job.num_cached_traces:
                 self.training_job.simulation_traces = self.training_job.simulation_traces[1:]
-            MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)
+            if self.save_to_metastore:
+                MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)
 
             # Update execution
             ts = time.time()
             self.exp_execution.timestamp = ts
             self.exp_execution.result = self.exp_result
-            MetastoreFacade.update_experiment_execution(experiment_execution=self.exp_execution,
-                                                        id=self.exp_execution.id)
+            if self.save_to_metastore:
+                MetastoreFacade.update_experiment_execution(experiment_execution=self.exp_execution,
+                                                            id=self.exp_execution.id)
 
         self.iter += 1
