@@ -1,8 +1,9 @@
+from typing import List, Optional, Tuple, Any
 import math
-from typing import List, Optional, Tuple
 import time
 import os
 import numpy as np
+import numpy.typing as npt
 import gymnasium as gym
 from csle_common.dao.simulation_config.simulation_env_config import SimulationEnvConfig
 from csle_common.dao.training.experiment_config import ExperimentConfig
@@ -15,6 +16,7 @@ from csle_common.dao.jobs.training_job_config import TrainingJobConfig
 from csle_common.dao.training.experiment_execution import ExperimentExecution
 from csle_common.dao.training.tabular_policy import TabularPolicy
 from csle_common.util.general_util import GeneralUtil
+from csle_common.dao.simulation_config.base_env import BaseEnv
 from csle_agents.agents.base.base_agent import BaseAgent
 import csle_agents.constants.constants as agents_constants
 
@@ -26,7 +28,8 @@ class QLearningAgent(BaseAgent):
 
     def __init__(self, simulation_env_config: SimulationEnvConfig,
                  experiment_config: ExperimentConfig,
-                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True):
+                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True,
+                 env: Optional[BaseEnv] = None):
         """
         Initializes the q-learning agent
 
@@ -34,14 +37,14 @@ class QLearningAgent(BaseAgent):
         :param experiment_config: the experiment configuration
         :param training_job: an existing training job to use (optional)
         :param save_to_metastore: boolean flag whether to save the execution to the metastore
+        :param env: the gym environment for training
         """
         super().__init__(simulation_env_config=simulation_env_config, emulation_env_config=None,
                          experiment_config=experiment_config)
         assert experiment_config.agent_type == AgentType.Q_LEARNING
         self.training_job = training_job
         self.save_to_metastore = save_to_metastore
-        self.env = gym.make(self.simulation_env_config.gym_env_name,
-                            config=self.simulation_env_config.simulation_env_input_config)
+        self.env = env
 
     def train(self) -> ExperimentExecution:
         """
@@ -65,12 +68,16 @@ class QLearningAgent(BaseAgent):
             exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = []
             exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = []
 
+        if self.env is None:
+            self.env = gym.make(self.simulation_env_config.gym_env_name,
+                                config=self.simulation_env_config.simulation_env_input_config)
+
         # Initialize training job
         if self.training_job is None:
             self.training_job = TrainingJobConfig(
                 simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                 progress_percentage=0, pid=pid, experiment_result=exp_result,
-                emulation_env_name=None, simulation_traces=[],
+                emulation_env_name="", simulation_traces=[],
                 num_cached_traces=0,
                 log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                 physical_host_ip=GeneralUtil.get_host_ip())
@@ -152,7 +159,8 @@ class QLearningAgent(BaseAgent):
         return [agents_constants.COMMON.EVAL_BATCH_SIZE, agents_constants.COMMON.CONFIDENCE_INTERVAL,
                 agents_constants.COMMON.RUNNING_AVERAGE, agents_constants.COMMON.GAMMA,
                 agents_constants.Q_LEARNING.EPSILON, agents_constants.Q_LEARNING.N,
-                agents_constants.Q_LEARNING.S, agents_constants.Q_LEARNING.A]
+                agents_constants.Q_LEARNING.S, agents_constants.Q_LEARNING.A,
+                agents_constants.Q_LEARNING.EPSILON_DECAY_RATE]
 
     def q_learning(self, exp_result: ExperimentResult, seed: int) -> ExperimentResult:
         """
@@ -167,11 +175,12 @@ class QLearningAgent(BaseAgent):
         A = self.experiment_config.hparams[agents_constants.Q_LEARNING.A].value
         epsilon = self.experiment_config.hparams[agents_constants.Q_LEARNING.EPSILON].value
         N = self.experiment_config.hparams[agents_constants.Q_LEARNING.N].value
+        epsilon_decay = self.experiment_config.hparams[agents_constants.Q_LEARNING.EPSILON_DECAY_RATE].value
         Logger.__call__().get_logger().info(f"Starting the q-learning algorithm, N:{N}, "
                                             f"num_states:{len(S)}, discount_factor: {discount_factor}, "
                                             f"num_actions: {len(A)}, epsilon: {epsilon}")
         avg_returns, running_avg_returns, initial_state_values, q_table, policy = self.train_q_learning(
-            A=A, S=S, gamma=discount_factor, N=N, epsilon=epsilon)
+            A=A, S=S, gamma=discount_factor, N=N, epsilon=epsilon, epsilon_decay=epsilon_decay)
         exp_result.all_metrics[seed][agents_constants.Q_LEARNING.INITIAL_STATE_VALUES] = initial_state_values
         exp_result.all_metrics[seed][agents_constants.COMMON.AVERAGE_RETURN] = avg_returns
         exp_result.all_metrics[seed][agents_constants.COMMON.RUNNING_AVERAGE_RETURN] = running_avg_returns
@@ -184,7 +193,8 @@ class QLearningAgent(BaseAgent):
         exp_result.policies[seed] = tabular_policy
         return exp_result
 
-    def train_q_learning(self, A: List, S: List, gamma: float = 0.8, N: int = 10000, epsilon: float = 0.2) \
+    def train_q_learning(self, A: List[int], S: List[int], gamma: float = 0.8, N: int = 10000, epsilon: float = 0.2,
+                         epsilon_decay: float = 1.0) \
             -> Tuple[List[float], List[float], List[float], List[List[float]], List[List[float]]]:
         """
         Runs the Q learning algorithm
@@ -194,68 +204,70 @@ class QLearningAgent(BaseAgent):
         :param gamma: the discount factor
         :param N: the number of iterations
         :param epsilon: the exploration parameter
+        :param epsilon_decay: the epsilon decay rate
         :return: the average returns, the running average returns, the initial state values, the q table, policy
         """
-        init_state_values = []
-        average_returns = []
-        running_average_returns = []
+        if self.env is None:
+            raise ValueError("Need to specify an environment to run Q-learning")
+        init_state_values: List[float] = []
+        average_returns: List[float] = []
+        running_average_returns: List[float] = []
 
-        Logger.__call__().get_logger().info("Starting Q Learning, gamma:{}, n_iter:{}, eps:{}".format(gamma, N,
-                                                                                                      epsilon))
+        Logger.__call__().get_logger().info(f"Starting Q Learning, gamma:{gamma}, n_iter:{N}, eps:{epsilon}")
         q_table = self.initialize_q_table(n_states=len(S), n_actions=len(A))
-        count_table = self.initialize_count_table(n_states=256, n_actions=5)
+        count_table = self.initialize_count_table(n_states=len(S), n_actions=len(A))
         steps = []
+        prog = 0.0
 
-        o = self.env.reset()
+        o, _ = self.env.reset()
         if self.simulation_env_config.gym_env_name in agents_constants.COMMON.STOPPING_ENVS:
             s = int(o[2])
         else:
-            s = o
+            s = int(o[0])
         for i in range(N):
             a = self.eps_greedy(q_table=q_table, A=A, s=s, epsilon=epsilon)
-            o, r, done, _ = self.env.step(a)
+            epsilon = epsilon * epsilon_decay
+            o, r, done, _, _ = self.env.step(a)
             if self.simulation_env_config.gym_env_name in agents_constants.COMMON.STOPPING_ENVS:
                 s_prime = int(o[2])
             else:
-                s_prime = o
-            q_table, count_table = self.q_learning_update(q_table=q_table, count_table=count_table,
-                                                          s=s, a=a, r=r, s_prime=s_prime, gamma=gamma)
-
+                s_prime = int(o[0])
+            q_table, count_table, alpha = self.q_learning_update(q_table=q_table, count_table=count_table,
+                                                                 s=s, a=a, r=r, s_prime=s_prime, gamma=gamma, done=done)
             if i % self.experiment_config.hparams[agents_constants.COMMON.EVAL_EVERY].value == 0:
                 steps.append(i)
-                state_val = np.sum(
-                    np.dot(
-                        np.array(list(map(lambda x: sum(q_table[x]), S))),
-                        np.array(
-                            self.simulation_env_config.initial_state_distribution_config.initial_state_distribution)))
                 prog = float(i / N)
-                init_state_values.append(state_val)
-
                 policy = self.create_policy_from_q_table(num_states=len(S), num_actions=len(A), q_table=q_table)
+                initial_state = int(np.random.choice(
+                    S, p=self.simulation_env_config.initial_state_distribution_config.initial_state_distribution))
+                state_val = q_table[initial_state][int(np.argmax(policy[initial_state]))]
+                init_state_values.append(state_val)
                 avg_return = self.evaluate_policy(policy=policy, eval_batch_size=self.experiment_config.hparams[
                     agents_constants.COMMON.EVAL_BATCH_SIZE].value)
                 average_returns.append(avg_return)
                 running_avg_J = ExperimentUtil.running_average(
                     average_returns, self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
                 running_average_returns.append(running_avg_J)
+                done = True
 
             if i % self.experiment_config.log_every == 0 or i == 0:
                 Logger.__call__().get_logger().info(
-                    "[Q-Learning] i:{}, progress:{}, V(s0):{}, J:{}, running_avg_J:{}".format(
-                        i, prog, state_val, avg_return, running_average_returns))
+                    f"[Q-Learning] i:{i}/{N}, progress:{prog}, V(s0):{state_val}, J:{avg_return}, "
+                    f"running_avg_J:{running_avg_J}, epsilon: {epsilon}, alpha: {alpha}")
 
             s = s_prime
             if done:
-                o = self.env.reset()
+                o, _ = self.env.reset()
                 if self.simulation_env_config.gym_env_name in agents_constants.COMMON.STOPPING_ENVS:
                     s = int(o[2])
                 else:
-                    s = o
+                    s = int(o[0])
 
         policy = self.create_policy_from_q_table(num_states=len(S), num_actions=len(A), q_table=q_table)
-        return average_returns, running_average_returns, init_state_values, q_table, policy
+        return (average_returns, running_average_returns, init_state_values, list(q_table.tolist()),
+                list(policy.tolist()))
 
-    def initialize_q_table(self, n_states: int = 256, n_actions: int = 5) -> np.ndarray:
+    def initialize_q_table(self, n_states: int = 256, n_actions: int = 5) -> npt.NDArray[Any]:
         """
         Initializes the Q table
 
@@ -263,10 +275,10 @@ class QLearningAgent(BaseAgent):
         :param n_actions: the number of actions in the MDP
         :return: the initialized Q table
         """
-        q_table = np.zeros((n_states, n_actions))
+        q_table = np.full((n_states, n_actions), -300.0, dtype=float)
         return q_table
 
-    def initialize_count_table(self, n_states: int = 256, n_actions: int = 5) -> np.ndarray:
+    def initialize_count_table(self, n_states: int = 256, n_actions: int = 5) -> npt.NDArray[Any]:
         """
         Initializes the count table
 
@@ -277,7 +289,7 @@ class QLearningAgent(BaseAgent):
         count_table = np.zeros((n_states, n_actions))
         return count_table
 
-    def eps_greedy(self, q_table: np.ndarray, A: List, s: int, epsilon: float = 0.2) -> int:
+    def eps_greedy(self, q_table: npt.NDArray[Any], A: List[int], s: int, epsilon: float = 0.2) -> int:
         """
         Selects an action according to the epsilon-greedy strategy
 
@@ -291,7 +303,7 @@ class QLearningAgent(BaseAgent):
             a = np.random.choice(A)
         else:
             a = np.argmax(q_table[s])
-        return a
+        return int(a)
 
     def step_size(self, n: int) -> float:
         """
@@ -302,8 +314,9 @@ class QLearningAgent(BaseAgent):
         """
         return float(1) / math.pow(n, 2 / 3)
 
-    def q_learning_update(self, q_table: np.ndarray, count_table: np.ndarray,
-                          s: int, a: int, r: float, s_prime: int, gamma: float) -> Tuple[np.ndarray, np.ndarray]:
+    def q_learning_update(self, q_table: npt.NDArray[Any], count_table: npt.NDArray[Any],
+                          s: int, a: int, r: float, s_prime: int, gamma: float, done: bool) \
+            -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
         """
         Watkin's Q-learning update
 
@@ -314,14 +327,19 @@ class QLearningAgent(BaseAgent):
         :param r: the reward
         :param s_prime: the next sampled state
         :param gamma: the discount factor
-        :return: the updated q table and updated count table
+        :param done: boolean flag indicating whether s_prime is terminal
+        :return: the updated q table and updated count table and the updated learning rate
         """
         count_table[s][a] = count_table[s_prime][a] + 1
         alpha = self.step_size(count_table[s][a])
-        q_table[s][a] = q_table[s][a] + alpha * ((r + gamma * np.max(q_table[s_prime])) - q_table[s][a])
-        return q_table, count_table
+        if not done:
+            q_table[s][a] = q_table[s][a] + alpha * ((r + gamma * np.max(q_table[s_prime])) - q_table[s][a])
+        else:
+            q_table[s][a] = q_table[s][a] + alpha * (r - q_table[s][a])
+        return q_table, count_table, alpha
 
-    def create_policy_from_q_table(self, num_states: int, num_actions: int, q_table: np.ndarray) -> np.ndarray:
+    def create_policy_from_q_table(self, num_states: int, num_actions: int, q_table: npt.NDArray[Any]) \
+            -> npt.NDArray[Any]:
         """
         Creates a tabular policy from a q table
 
@@ -339,7 +357,7 @@ class QLearningAgent(BaseAgent):
             policy[s][best_action] = 1.0
         return policy
 
-    def evaluate_policy(self, policy: np.ndarray, eval_batch_size: int) -> float:
+    def evaluate_policy(self, policy: npt.NDArray[Any], eval_batch_size: int) -> float:
         """
         Evalutes a tabular policy
 
@@ -347,13 +365,31 @@ class QLearningAgent(BaseAgent):
         :param eval_batch_size: the batch size
         :return: None
         """
+        if self.env is None:
+            raise ValueError("Need to specify an environment to run policy evaluation")
         returns = []
         for i in range(eval_batch_size):
             done = False
-            s = self.env.reset()
+            o, _ = self.env.reset()
+            if self.simulation_env_config.gym_env_name in agents_constants.COMMON.STOPPING_ENVS:
+                s = int(o[2])
+            else:
+                s = int(o[0])
             R = 0
+            t = 0
             while not done:
-                s, r, done, info = self.env.step(policy)
+                t += 1
+                if self.simulation_env_config.gym_env_name in agents_constants.COMMON.STOPPING_ENVS:
+                    o, r, done, _, info = self.env.step(policy)
+                else:
+                    a = int(np.argmax(policy[s]))
+                    if t == 1:
+                        print(f"initial action: {a}")
+                    o, r, done, _, info = self.env.step(a)
+                if self.simulation_env_config.gym_env_name in agents_constants.COMMON.STOPPING_ENVS:
+                    s = int(o[2])
+                else:
+                    s = int(o[0])
                 R += r
             returns.append(R)
         avg_return = np.mean(returns)
