@@ -1,5 +1,5 @@
+from typing import List, Optional, Tuple, Any
 import math
-from typing import List, Optional, Tuple
 import time
 import os
 import numpy as np
@@ -15,6 +15,7 @@ from csle_common.dao.jobs.training_job_config import TrainingJobConfig
 from csle_common.dao.training.experiment_execution import ExperimentExecution
 from csle_common.dao.training.alpha_vectors_policy import AlphaVectorsPolicy
 from csle_common.util.general_util import GeneralUtil
+from csle_common.dao.simulation_config.base_env import BaseEnv
 from csle_agents.agents.base.base_agent import BaseAgent
 import csle_agents.constants.constants as agents_constants
 from scipy.optimize import linprog
@@ -28,7 +29,8 @@ class SondikVIAgent(BaseAgent):
 
     def __init__(self, simulation_env_config: SimulationEnvConfig,
                  experiment_config: ExperimentConfig,
-                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True):
+                 training_job: Optional[TrainingJobConfig] = None, save_to_metastore: bool = True,
+                 env: Optional[BaseEnv] = None):
         """
         Initializes the Sondik-VI agent
 
@@ -36,14 +38,14 @@ class SondikVIAgent(BaseAgent):
         :param experiment_config: the experiment configuration
         :param training_job: an existing training job to use (optional)
         :param save_to_metastore: boolean flag whether to save the execution to the metastore
+        :param env: the gym environment for training
         """
         super().__init__(simulation_env_config=simulation_env_config, emulation_env_config=None,
                          experiment_config=experiment_config)
         assert experiment_config.agent_type == AgentType.SONDIK_VALUE_ITERATION
         self.training_job = training_job
         self.save_to_metastore = save_to_metastore
-        self.env = gym.make(self.simulation_env_config.gym_env_name,
-                            config=self.simulation_env_config.simulation_env_input_config)
+        self.env = env
 
     def train(self) -> ExperimentExecution:
         """
@@ -70,6 +72,10 @@ class SondikVIAgent(BaseAgent):
             exp_result.all_metrics[seed][agents_constants.SONDIK_VI.NUM_ALPHA_VECTORS] = []
             exp_result.all_metrics[seed][agents_constants.SONDIK_VI.INITIAL_BELIEF_VALUES] = []
 
+        if self.env is None:
+            self.env = gym.make(self.simulation_env_config.gym_env_name,
+                                config=self.simulation_env_config.simulation_env_input_config)
+
         # Initialize training job
         if self.training_job is None:
             self.training_job = TrainingJobConfig(
@@ -222,7 +228,7 @@ class SondikVIAgent(BaseAgent):
 
     def sondik_vi(self, P, Z, R, T, gamma, n_states, n_actions, n_obs, b0, eval_batch_size: int,
                   use_pruning: bool = True) \
-            -> Tuple[List[float], List[float], List[float], List[float], List[float]]:
+            -> Tuple[List[Any], List[int], List[float], List[float], List[float]]:
         """
 
         :param P: The transition probability matrix
@@ -243,11 +249,11 @@ class SondikVIAgent(BaseAgent):
         # a set of values
         alepth_t_plus_1.add(zero_alpha_vec)
         first = True
-        num_alpha_vectors = []
+        num_alpha_vectors: List[int] = []
         num_alpha_vectors.append(len(alepth_t_plus_1))
-        initial_values = []
-        average_returns = []
-        average_running_returns = []
+        initial_values: List[float] = []
+        average_returns: List[float] = []
+        average_running_returns: List[float] = []
 
         # Backward induction
         for t in range(T):
@@ -319,7 +325,7 @@ class SondikVIAgent(BaseAgent):
                 if v > max_v:
                     max_v = v
             initial_values.append(max_v)
-            avg_R = -1
+            avg_R = -1.0
             if len(average_returns) > 0:
                 avg_R = average_returns[-1]
             alpha_vec_policy = AlphaVectorsPolicy(
@@ -335,8 +341,8 @@ class SondikVIAgent(BaseAgent):
             running_avg_J = ExperimentUtil.running_average(
                 average_returns, self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value)
             average_running_returns.append(running_avg_J)
-
-        return (list(map(lambda x: x[1], list(aleph_t))), num_alpha_vectors, initial_values, average_returns,
+        vectors: List[Any] = list(map(lambda x: x[1], list(aleph_t)))
+        return (vectors, num_alpha_vectors, initial_values, average_returns,
                 average_running_returns)
 
     def prune(self, n_states, aleph):
@@ -422,16 +428,18 @@ class SondikVIAgent(BaseAgent):
         :param eval_batch_size: the batch size
         :return: None
         """
+        if self.env is None:
+            raise ValueError("An environment must be specified to run policy evaluation")
         returns = []
         for i in range(eval_batch_size):
             done = False
-            o = self.env.reset()
+            o, _ = self.env.reset()
             R = 0
             while not done:
                 b1 = o[1]
                 b = [1 - b1, b1, 0]
                 a = policy.action(b)
-                o, r, done, info = self.env.step(a)
+                o, r, done, _, info = self.env.step(a)
                 R += r
             returns.append(R)
         avg_return = np.mean(returns)
