@@ -20,6 +20,7 @@ from csle_common.dao.training.mixed_multi_threshold_stopping_policy import Mixed
 from csle_common.dao.training.policy import Policy
 import csle_common.constants.constants as constants
 from csle_common.util.general_util import GeneralUtil
+from csle_common.dao.simulation_config.base_env import BaseEnv
 from csle_agents.agents.base.base_agent import BaseAgent
 from csle_agents.agents.t_spsa.t_spsa_agent import TSPSAAgent
 import csle_agents.constants.constants as agents_constants
@@ -28,7 +29,8 @@ import gym_csle_stopping_game.constants.constants as env_constants
 
 class TFPAgent(BaseAgent):
     """
-    RL Agent implementing the T-FP algorithm from TODO
+    RL Agent implementing the T-FP algorithm from (Hammar & Stadler '23 - Learning Near-Optimal Intrusion Responses
+    Against Dynamic Attackers)
     """
 
     def __init__(self, defender_simulation_env_config: SimulationEnvConfig,
@@ -125,10 +127,13 @@ class TFPAgent(BaseAgent):
                 exp_result.all_metrics[seed][env_constants.ENV_METRICS.STOP + f"_running_average_{l}"] = []
 
         if self.training_job is None:
+            emulation_name = ""
+            if self.emulation_env_config is not None:
+                emulation_name = self.emulation_env_config.name
             self.training_job = TrainingJobConfig(
                 simulation_env_name=self.simulation_env_config.name, experiment_config=self.experiment_config,
                 experiment_result=exp_result, progress_percentage=0, pid=pid,
-                emulation_env_name=self.emulation_env_config.name, simulation_traces=[],
+                emulation_env_name=emulation_name, simulation_traces=[],
                 num_cached_traces=agents_constants.COMMON.NUM_CACHED_SIMULATION_TRACES,
                 log_file_path=Logger.__call__().get_log_file_path(), descr=descr,
                 physical_host_ip=GeneralUtil.get_host_ip())
@@ -140,7 +145,7 @@ class TFPAgent(BaseAgent):
             self.training_job.experiment_result = exp_result
             MetastoreFacade.update_training_job(training_job=self.training_job, id=self.training_job.id)
         config = self.simulation_env_config.simulation_env_input_config
-        env = gym.make(self.simulation_env_config.gym_env_name, config=config)
+        env: BaseEnv = gym.make(self.simulation_env_config.gym_env_name, config=config)
         for seed in self.experiment_config.random_seeds:
             ExperimentUtil.set_seed(seed)
             exp_result = self.t_fp(exp_result=exp_result, seed=seed, env=env, training_job=self.training_job,
@@ -162,31 +167,47 @@ class TFPAgent(BaseAgent):
                 seed_values = []
                 for seed_idx in range(len(self.experiment_config.random_seeds)):
                     seed_values.append(value_vectors[seed_idx][i])
-                avg_metrics.append(ExperimentUtil.mean_confidence_interval(data=seed_values, confidence=confidence)[0])
-                std_metrics.append(ExperimentUtil.mean_confidence_interval(data=seed_values, confidence=confidence)[1])
+                try:
+                    avg_metrics.append(
+                        ExperimentUtil.mean_confidence_interval(data=seed_values, confidence=confidence)[0])
+                    std_metrics.append(
+                        ExperimentUtil.mean_confidence_interval(data=seed_values, confidence=confidence)[1])
+                except Exception:
+                    pass
             exp_result.avg_metrics[metric] = avg_metrics
             exp_result.std_metrics[metric] = std_metrics
 
         ts = time.time()
-        emulation_name = None
+        emulation_name = ""
         if self.emulation_env_config is not None:
             emulation_name = self.emulation_env_config.name
         simulation_name = self.simulation_env_config.name
         exp_execution = ExperimentExecution(result=exp_result, config=self.experiment_config, timestamp=ts,
                                             emulation_name=emulation_name, simulation_name=simulation_name,
                                             descr=descr, log_file_path=self.training_job.log_file_path)
-        traces = env.get_traces()
+        traces = env.unwrapped.get_traces()
         if len(traces) > 0:
             MetastoreFacade.save_simulation_trace(traces[-1])
         MetastoreFacade.remove_training_job(self.training_job)
         return exp_execution
 
-    def t_fp(self, exp_result: ExperimentResult, seed: int, env: gym.Env,
+    def t_fp(self, exp_result: ExperimentResult, seed: int, env: BaseEnv,
              training_job: TrainingJobConfig, random_seeds: List[int]):
+        """
+        Runs the T-FP algorithm (Hammar, Stadler 2023)
+
+        :param exp_result: the experiment result
+        :param seed: the seed for the experiment
+        :param env: environment for evaluation
+        :param training_job: the training job for the evaluation
+        :param random_seeds: the random seeds for the evaluation
+        :return: the experiment result
+        """
 
         # Initialize policies
         defender_policy = MixedMultiThresholdStoppingPolicy(
-            Theta=np.zeros((self.experiment_config.hparams[constants.T_SPSA.L].value,)).tolist(),
+            defender_Theta=np.zeros((self.experiment_config.hparams[constants.T_SPSA.L].value, 2, 1)).tolist(),
+            attacker_Theta=[],
             simulation_name=self.defender_simulation_env_config.name,
             states=self.defender_simulation_env_config.state_space_config.states,
             player_type=PlayerType.DEFENDER,
@@ -196,7 +217,8 @@ class TFPAgent(BaseAgent):
             experiment_config=self.defender_experiment_config, avg_R=-1,
             agent_type=AgentType.T_FP)
         attacker_policy = MixedMultiThresholdStoppingPolicy(
-            Theta=np.zeros((2, self.experiment_config.hparams[constants.T_SPSA.L].value)).tolist(),
+            attacker_Theta=np.zeros((2, self.experiment_config.hparams[constants.T_SPSA.L].value, 2, 1)).tolist(),
+            defender_Theta=[],
             simulation_name=self.attacker_simulation_env_config.name,
             states=self.attacker_simulation_env_config.state_space_config.states,
             player_type=PlayerType.ATTACKER,
@@ -206,32 +228,16 @@ class TFPAgent(BaseAgent):
             experiment_config=self.attacker_experiment_config, avg_R=-1,
             agent_type=AgentType.T_FP, opponent_strategy=defender_policy)
 
-        # Update average policies with initial thresholds
-        # initial_attacker_thresholds = [
-        #     [[0]*self.attacker_experiment_config.hparams[constants.T_SPSA.L].value,
-        #      [0]*self.attacker_experiment_config.hparams[constants.T_SPSA.L].value
-        #      ]*10
-        # ]
-        initial_attacker_thresholds = []
-        initial_defender_thresholds = []
+        initial_attacker_thresholds: List[List[List[float]]] = []
+        initial_defender_thresholds: List[List[float]] = []
         initial_attacker_thresholds.append(
-            [[0] * self.attacker_experiment_config.hparams[constants.T_SPSA.L].value,
-             [1] * self.attacker_experiment_config.hparams[constants.T_SPSA.L].value
-             ]
-        )
-        initial_defender_thresholds.append(
-            [0] * self.attacker_experiment_config.hparams[constants.T_SPSA.L].value)
-        # for i in range(1):
-        #     initial_attacker_thresholds.append(
-        #         [[round(random.uniform(0,1),2)]*self.attacker_experiment_config.hparams[constants.T_SPSA.L].value,
-        #          [round(random.uniform(0,1),2)]*self.attacker_experiment_config.hparams[constants.T_SPSA.L].value
-        #          ]
-        #     )
-        #     initial_defender_thresholds.append(
-        #         [round(random.uniform(0,1),2)]*self.attacker_experiment_config.hparams[constants.T_SPSA.L].value)
+            [[0.0] * self.attacker_experiment_config.hparams[constants.T_SPSA.L].value,
+             [1.0] * self.attacker_experiment_config.hparams[constants.T_SPSA.L].value
+             ])
+        initial_defender_thresholds.append([0.0] * self.attacker_experiment_config.hparams[constants.T_SPSA.L].value)
 
-        attacker_policy.update_Theta(new_thresholds=initial_attacker_thresholds)
-        defender_policy.update_Theta(new_thresholds=initial_defender_thresholds)
+        attacker_policy._update_Theta_attacker(new_thresholds=initial_attacker_thresholds)
+        defender_policy._update_Theta_defender(new_thresholds=initial_defender_thresholds)
         attacker_policy.opponent_strategy = defender_policy
 
         for i in range(self.experiment_config.hparams[agents_constants.T_FP.N_2].value):
@@ -255,23 +261,11 @@ class TFPAgent(BaseAgent):
             defender_val = round(defender_metrics[env_constants.ENV_METRICS.RETURN], 3)
             val = round(strategy_profile_metrics[env_constants.ENV_METRICS.RETURN], 3)
 
-            attacker_policy.update_Theta(new_thresholds=[attacker_thresholds])
-            defender_policy.update_Theta(new_thresholds=[defender_thresholds])
+            attacker_policy._update_Theta_attacker(new_thresholds=[attacker_thresholds])
+            defender_policy._update_Theta_defender(new_thresholds=[defender_thresholds])
             val_attacker_exp = attacker_val
             val_defender_exp = defender_val
 
-            # # Update empirical strategies
-            # if attacker_val > -defender_val:
-            #     attacker_policy.update_Theta(new_thresholds=[attacker_thresholds])
-            #     val_attacker_exp= attacker_val
-            # else:
-            #     val_attacker_exp = -defender_val
-            # if defender_val > -attacker_val:
-            #     defender_policy.update_Theta(new_thresholds=[defender_thresholds])
-            #     val_defender_exp=defender_val
-            # else:
-            #     val_defender_exp = -attacker_val
-
             attacker_policy.opponent_strategy = defender_policy
 
             # Log rewards
@@ -300,8 +294,10 @@ class TFPAgent(BaseAgent):
                     self.experiment_config.hparams[agents_constants.COMMON.RUNNING_AVERAGE].value))
 
             # Log thresholds
-            exp_result.all_metrics[seed][agents_constants.T_FP.ATTACKER_THRESHOLDS].append(attacker_policy.Theta)
-            exp_result.all_metrics[seed][agents_constants.T_FP.DEFENDER_THRESHOLDS].append(defender_policy.Theta)
+            exp_result.all_metrics[seed][agents_constants.T_FP.ATTACKER_THRESHOLDS].append(
+                attacker_policy.attacker_Theta)
+            exp_result.all_metrics[seed][agents_constants.T_FP.DEFENDER_THRESHOLDS].append(
+                defender_policy.defender_Theta)
 
             # Log stop distributions
             for k, v in attacker_policy.stop_distributions().items():
@@ -376,6 +372,7 @@ class TFPAgent(BaseAgent):
                 progress = round(iterations_done / total_iterations, 2)
                 training_job.progress_percentage = progress
                 MetastoreFacade.update_training_job(training_job=training_job, id=training_job.id)
+        return exp_result
 
     def evaluate_defender_policy(self, defender_thresholds: List[float],
                                  attacker_strategy: MixedMultiThresholdStoppingPolicy) -> Dict[str, Union[float, int]]:
@@ -411,10 +408,10 @@ class TFPAgent(BaseAgent):
         :return: the average reward
         """
         self.attacker_simulation_env_config.simulation_env_input_config.defender_strategy = defender_strategy
-        env = gym.make(self.attacker_simulation_env_config.gym_env_name,
-                       config=self.attacker_simulation_env_config.simulation_env_input_config)
-        env.set_model(attacker_strategy)
-        attacker_strategy.opponent_strategy = env.static_defender_strategy
+        env: BaseEnv = gym.make(self.attacker_simulation_env_config.gym_env_name,
+                                config=self.attacker_simulation_env_config.simulation_env_input_config)
+        env.unwrapped.set_model(attacker_strategy)
+        attacker_strategy.opponent_strategy = env.unwrapped.static_defender_strategy
         return self._eval_env(
             env=env, policy=attacker_strategy,
             num_iterations=self.experiment_config.hparams[
@@ -440,17 +437,17 @@ class TFPAgent(BaseAgent):
                 self.attacker_experiment_config.player_idx].actions, experiment_config=self.attacker_experiment_config,
             avg_R=-1, agent_type=AgentType.NONE)
         self.attacker_simulation_env_config.simulation_env_input_config.defender_strategy = defender_strategy
-        env = gym.make(self.attacker_simulation_env_config.gym_env_name,
-                       config=self.attacker_simulation_env_config.simulation_env_input_config)
-        env.set_model(attacker_strategy)
-        attacker_policy.opponent_strategy = env.static_defender_strategy
+        env: BaseEnv = gym.make(self.attacker_simulation_env_config.gym_env_name,
+                                config=self.attacker_simulation_env_config.simulation_env_input_config)
+        env.unwrapped.set_model(attacker_strategy)
+        attacker_policy.opponent_strategy = env.unwrapped.static_defender_strategy
         return self._eval_env(
             env=env, policy=attacker_policy,
             num_iterations=self.experiment_config.hparams[
                 agents_constants.T_FP.BEST_RESPONSE_EVALUATION_ITERATIONS].value)
 
     def defender_best_response(self, seed: int, attacker_strategy: MixedMultiThresholdStoppingPolicy) \
-            -> Tuple[List, float]:
+            -> Tuple[List[float], float]:
         """
         Learns a best response for the defender against a given attacker strategy
 
@@ -474,23 +471,24 @@ class TFPAgent(BaseAgent):
         val = experiment_execution.result.avg_metrics[agents_constants.COMMON.RUNNING_AVERAGE_RETURN][-1]
         return thresholds, val
 
-    def _eval_env(self, env: gym.Env, policy: Policy, num_iterations: int) -> Dict[str, Union[float, int]]:
+    def _eval_env(self, env: BaseEnv, policy: Policy, num_iterations: int) -> Dict[str, Union[float, int]]:
         """
+        Evaluates a given policy
 
         :param env: the environment to use for evaluation
         :param policy: the policy to evaluate
         :param num_iterations: number of iterations to evaluate
         :return: the average reward
         """
-        metrics = {}
+        metrics: Dict[str, List[Union[float, int]]] = {}
         for j in range(num_iterations):
             done = False
-            o = env.reset()
+            o, _ = env.reset()
             J = 0
             t = 1
             while not done and t <= self.experiment_config.hparams[agents_constants.COMMON.MAX_ENV_STEPS].value:
                 a = policy.action(o=o)
-                o, r, done, info = env.step(a)
+                o, r, done, _, info = env.step(a)
                 J += r
                 t += 1
             metrics = TSPSAAgent.update_metrics(metrics=metrics, info=info)
@@ -529,8 +527,7 @@ class TFPAgent(BaseAgent):
         return avg_metrics
 
     def attacker_best_response(self, seed: int, defender_strategy: MixedMultiThresholdStoppingPolicy,
-                               attacker_strategy: MixedMultiThresholdStoppingPolicy) \
-            -> Tuple[List, float]:
+                               attacker_strategy: MixedMultiThresholdStoppingPolicy) -> Tuple[List[List[float]], float]:
         """
         Learns a threshold best response strategy for the attacker against a given defender strategy
 
@@ -542,9 +539,9 @@ class TFPAgent(BaseAgent):
         self.attacker_experiment_config.random_seeds = [seed]
         self.attacker_experiment_config.output_dir = str(self.root_output_dir)
         self.attacker_simulation_env_config.simulation_env_input_config.defender_strategy = defender_strategy
-        env = gym.make(self.attacker_simulation_env_config.gym_env_name,
-                       config=self.attacker_simulation_env_config.simulation_env_input_config)
-        env.set_model(attacker_strategy)
+        env: BaseEnv = gym.make(self.attacker_simulation_env_config.gym_env_name,
+                                config=self.attacker_simulation_env_config.simulation_env_input_config)
+        env.unwrapped.set_model(attacker_strategy)
         agent = TSPSAAgent(emulation_env_config=self.emulation_env_config,
                            simulation_env_config=self.attacker_simulation_env_config,
                            experiment_config=self.attacker_experiment_config,
@@ -572,7 +569,8 @@ class TFPAgent(BaseAgent):
                 agents_constants.T_FP.N_2, constants.T_SPSA.GRADIENT_BATCH_SIZE,
                 agents_constants.COMMON.CONFIDENCE_INTERVAL, agents_constants.COMMON.RUNNING_AVERAGE,
                 agents_constants.T_FP.BEST_RESPONSE_EVALUATION_ITERATIONS,
-                agents_constants.T_FP.EQUILIBRIUM_STRATEGIES_EVALUATION_ITERATIONS]
+                agents_constants.T_FP.EQUILIBRIUM_STRATEGIES_EVALUATION_ITERATIONS,
+                constants.T_SPSA.POLICY_TYPE, constants.T_SPSA.OBJECTIVE_TYPE]
 
     @staticmethod
     def exploitability(attacker_val: float, defender_val: float) -> float:
@@ -598,6 +596,8 @@ class TFPAgent(BaseAgent):
             constants.T_SPSA.c: self.experiment_config.hparams[constants.T_SPSA.c],
             constants.T_SPSA.a: self.experiment_config.hparams[constants.T_SPSA.a],
             constants.T_SPSA.A: self.experiment_config.hparams[constants.T_SPSA.A],
+            constants.T_SPSA.POLICY_TYPE: self.experiment_config.hparams[constants.T_SPSA.POLICY_TYPE],
+            constants.T_SPSA.OBJECTIVE_TYPE: self.experiment_config.hparams[constants.T_SPSA.OBJECTIVE_TYPE],
             constants.T_SPSA.LAMBDA: self.experiment_config.hparams[constants.T_SPSA.LAMBDA],
             constants.T_SPSA.EPSILON: self.experiment_config.hparams[constants.T_SPSA.EPSILON],
             constants.T_SPSA.L: self.experiment_config.hparams[constants.T_SPSA.L],
@@ -634,6 +634,8 @@ class TFPAgent(BaseAgent):
             constants.T_SPSA.c: self.experiment_config.hparams[constants.T_SPSA.c],
             constants.T_SPSA.a: self.experiment_config.hparams[constants.T_SPSA.a],
             constants.T_SPSA.A: self.experiment_config.hparams[constants.T_SPSA.A],
+            constants.T_SPSA.POLICY_TYPE: self.experiment_config.hparams[constants.T_SPSA.POLICY_TYPE],
+            constants.T_SPSA.OBJECTIVE_TYPE: self.experiment_config.hparams[constants.T_SPSA.OBJECTIVE_TYPE],
             constants.T_SPSA.LAMBDA: self.experiment_config.hparams[constants.T_SPSA.LAMBDA],
             constants.T_SPSA.EPSILON: self.experiment_config.hparams[constants.T_SPSA.EPSILON],
             constants.T_SPSA.L: self.experiment_config.hparams[constants.T_SPSA.L],
@@ -683,9 +685,9 @@ class TFPAgent(BaseAgent):
         """
         if len(x) >= N:
             y = np.copy(x)
-            y[N - 1:] = np.convolve(x, np.ones((N, )) / N, mode='valid')
+            y[N - 1:] = np.convolve(x, np.ones((N,)) / N, mode='valid')
         else:
             N = len(x)
             y = np.copy(x)
-            y[N - 1:] = np.convolve(x, np.ones((N, )) / N, mode='valid')
-        return y.tolist()
+            y[N - 1:] = np.convolve(x, np.ones((N,)) / N, mode='valid')
+        return list(y.tolist())
